#include <stdio.h>
#include <stdlib.h>
#include <string.h> // memcpy
#include <unistd.h>
#include <termios.h>

#include <glib-object.h>
#include <glib/gprintf.h>
#include <dbus/dbus-glib.h>

#include "controller-term.h"

/**
 * SECTION: ControllerTerm
 * @title: ControllerTerm
 *
 * Missing Documentation about a controller term.
 *
 **/

/* static functions, boilerplate code */
static void ui_controller_term_init (UiControllerTerm *object);
static void ui_controller_term_class_init (UiControllerTermClass * controller_term_class);
static GObject *ui_controller_term_constructor (GType gtype,
    guint n_properties, GObjectConstructParam *properties);
static void ui_controller_term_finalize (GObject *object);
static void ui_controller_term_dispose (GObject *object);
static void ui_controller_term_get_property (GObject *object,
    guint property_id, GValue *value, GParamSpec *param_spec);
static void ui_controller_term_set_property (GObject *object, guint property_id,
    const GValue *value, GParamSpec *param_spec);

/* static functions, local */
static gboolean ui_controller_term_handle_stdin (GIOChannel *channel,
    GIOCondition condition, gpointer data);
 
static void ui_controller_term_add_stdin_source (UiControllerTerm *controller_term,
    GError **error);
static gboolean ui_controller_term_setup_terminal (UiControllerTerm *controller_term,
    GError **error);
static void ui_controller_term_restore_terminal (UiControllerTerm *controller_term,
    GError **error);

/* D-Bus */

#include "controller-term-server-stub.h"

/* static functions, interface implementation */
static gboolean ui_controller_term_configure_impl (UiController *controller,
    const gchar *filename, GError **error);
static gboolean ui_controller_term_connect_with_main_loop_impl (UiController *controller,
    GMainLoop *main_loop, GMainContext *main_context, GError **error);
static gboolean ui_controller_term_register_with_dbus_impl (UiController *controller,
    const gchar *object_path, DBusGConnection * bus, GError **error);

/* static objects */
static GObjectClass *parent_class = NULL;
static guint signals[UI_CONTROLLER_TERM_SIGNAL_COUNT] = { 0 };

/* private structure */
typedef struct _UiControllerTermPrivate UiControllerTermPrivate;

struct _UiControllerTermPrivate
{
  gchar *id;
  GMainLoop *main_loop;
  GMainContext *main_context;
  guint32 source_tag;
  struct termios new_term;
  struct termios old_term;
};

#define UI_CONTROLLER_TERM_GET_PRIVATE(obj) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), UI_TYPE_CONTROLLER_TERM, \
                                UiControllerTermPrivate))

/**
 * ui_controller_term_get_type:
 * @returns: a #GType
 *
 * boilerplate function, defines the type returned by UI_TYPE_CONTROLLER_TERM
 **/
GType
ui_controller_term_get_type (void)
{
  static GType ui_controller_term_type = 0;
  if (G_UNLIKELY (ui_controller_term_type == 0))
  {
    static const GTypeInfo info = {
      sizeof (UiControllerTermClass), /* class_size */
      (GBaseInitFunc) NULL,           /* base_init */
      (GBaseFinalizeFunc) NULL,       /* base_finalize */
      (GClassInitFunc) ui_controller_term_class_init, /* class_init */
      (GClassFinalizeFunc) NULL,      /* class_finalize */
      NULL,                           /* class_data */
      sizeof (UiControllerTerm),      /* instance_size */
      0,                              /* n_preallocs */
      (GInstanceInitFunc) ui_controller_term_init,  /* instance_init */
      NULL                            /* value_table */
    };

    ui_controller_term_type = g_type_register_static (
        UI_TYPE_CONTROLLER,           /* parent type */
        "UiControllerTerm",           /* name of the new type */
        &info,                        /* GTypeInfo for this type */
        0                             /* Flags (e.g abstract) */
        );
  }
  return ui_controller_term_type;
}

/**
 * ui_controller_term_init:
 * @ui_controller: A #UiControllerTerm
 *
 * boilerplate function, initializes the #UiControllerTerm
 **/
static void
ui_controller_term_init (UiControllerTerm *ui_controller_term)
{
  UiControllerTermPrivate *priv = UI_CONTROLLER_TERM_GET_PRIVATE (ui_controller_term);

  priv->id = g_strdup ("");
}

/**
 * ui_controller_term_class_init:
 * @ui_controller_term_class: A #UiControllerTermClass
 *
 * boilerplate function, initializes the #UiControllerTermClass
 **/
static void
ui_controller_term_class_init (UiControllerTermClass * ui_controller_term_class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (ui_controller_term_class);

  /* add private section */
  g_type_class_add_private (ui_controller_term_class, sizeof (UiControllerTermPrivate));

  /* define the parent class */
  parent_class = g_type_class_peek_parent (ui_controller_term_class);

  /* assign boilerplate functions */
  object_class->dispose = ui_controller_term_dispose;
  object_class->finalize = ui_controller_term_finalize;
  object_class->constructor = ui_controller_term_constructor;
  object_class->get_property = ui_controller_term_get_property;
  object_class->set_property = ui_controller_term_set_property;

  /* assign interface functions */
  UiControllerClass *ui_controller_class = UI_CONTROLLER_CLASS (ui_controller_term_class);
  ui_controller_class->configure = ui_controller_term_configure_impl;
  ui_controller_class->connect_with_main_loop
      = ui_controller_term_connect_with_main_loop_impl;
  ui_controller_class->register_with_dbus = ui_controller_term_register_with_dbus_impl;

  /* create properties */
  g_object_class_install_property (object_class,
      UI_CONTROLLER_TERM_PROPERTY_ID,
      g_param_spec_string ("id", "Id", "Id", NULL, G_PARAM_READWRITE));

  /* create signals */
  signals[UI_CONTROLLER_TERM_SIGNAL_KEY_UP] = g_signal_new (
      UI_CONTROLLER_TERM_SIGNAL_KEY_UP_STR,
      G_OBJECT_CLASS_TYPE (ui_controller_term_class),
      G_SIGNAL_RUN_LAST,
      0, /* No offset */
      NULL, /* No GSignalAccumulator to use */
      NULL, /* No user-data passed to the accumulator */
      g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

  /* binding to GLib/D-Bus */
  dbus_g_object_type_install_info (UI_TYPE_CONTROLLER_TERM,
      &dbus_glib_ui_controller_term_object_info);
}

/**
 * ui_controller_term_dispose:
 * @object:
 *
 * boilerplate function.
 **/
void
ui_controller_term_dispose (GObject *object)
{
  UiControllerTerm *self = UI_CONTROLLER_TERM (object);
  UiControllerTermPrivate *priv = UI_CONTROLLER_TERM_GET_PRIVATE (self);

  ui_controller_term_restore_terminal (self, NULL);

  /* release any references held by the object here */
  g_free (priv->id);

  if (G_OBJECT_CLASS (parent_class)->dispose)
    G_OBJECT_CLASS (parent_class)->dispose (object);
}

/**
 * ui_controller_term_finalize:
 * @object:
 *
 * boilerplate function.
 **/
void
ui_controller_term_finalize (GObject *object)
{
  if (G_OBJECT_CLASS (parent_class)->finalize)
    G_OBJECT_CLASS (parent_class)->finalize (object);
}

/**
 * ui_controller_term_constructor:
 * @gtype:
 * @n_properties:
 * @properties:
 *
 * boilerplate function.
 **/
static GObject *
ui_controller_term_constructor (GType gtype, guint n_properties,
    GObjectConstructParam *properties)
{
  GObjectClass *object_class = G_OBJECT_CLASS (parent_class);
  GObject *object = object_class->constructor (gtype, n_properties, properties);

  /* TODO Update object state / constructor properties */

  return object;
}

/**
 * ui_controller_term_get_property:
 * @object:
 * @property_id:
 * @value:
 * @param_spec:
 *
 * boilerplate function.
 **/
static void
ui_controller_term_get_property (GObject *object, guint property_id,
    GValue *value, GParamSpec *param_spec)
{
  g_return_if_fail (UI_IS_CONTROLLER_TERM (object));

  UiControllerTerm *controller_term = UI_CONTROLLER_TERM (object);
  UiControllerTermPrivate *priv = UI_CONTROLLER_TERM_GET_PRIVATE (controller_term);

  switch (property_id)
  {
    case UI_CONTROLLER_TERM_PROPERTY_ID:
      g_value_set_string (value, priv->id);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, param_spec);
      break;
  }
}

/**
 * ui_controller_term_set_property:
 * @object:
 * @property_id:
 * @value:
 * @param_spec:
 *
 * boilerplate function.
 **/
static void
ui_controller_term_set_property (GObject *object, guint property_id,
    const GValue *value, GParamSpec *param_spec)
{
  g_return_if_fail (UI_IS_CONTROLLER_TERM (object));

  UiControllerTerm *controller_term = UI_CONTROLLER_TERM (object);
  UiControllerTermPrivate *priv = UI_CONTROLLER_TERM_GET_PRIVATE (controller_term);

  switch (property_id)
  {
    case UI_CONTROLLER_TERM_PROPERTY_ID:
      g_free (priv->id);
      priv->id = g_value_dup_string (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, param_spec);
      break;
  };
}

/**
 * ui_controller_term_new:
 * @id:
 * @returns: A new UiController
 *
 * The user is expected to delete the controller with ui_controller_free when
 * it is no longer needed.
 **/
UiController *
ui_controller_term_new (gchar *id)
{
  g_return_val_if_fail (id != NULL, NULL);

  UiController *controller = g_object_new (UI_TYPE_CONTROLLER_TERM,
      "id", id,
      NULL);

  return controller;
}

/**
 * ui_controller_term_setup_terminal:
 * @controller_term:
 * @error:
 * @returns: %TRUE if the call was successful, %FALSE otherwise
 *
 * This function configures the terminal entry.
 **/
static gboolean
ui_controller_term_setup_terminal (UiControllerTerm *controller_term, GError **error)
{
  UiControllerTermPrivate *priv = UI_CONTROLLER_TERM_GET_PRIVATE (controller_term);
  
  if (isatty (STDIN_FILENO) == 0)
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_PROGRAMMING,
        "%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_setup_terminal",
        "standard input is not a terminal device");
    return FALSE;
  }
  // save the current settings, so they can be restored upon terminating the program.
  if (tcgetattr (STDIN_FILENO, &(priv->new_term)) < 0)
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_PROGRAMMING,
        "%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_setup_terminal",
        "could not get term attributes");
    return FALSE;
  }
  memcpy (&(priv->old_term), &(priv->new_term), sizeof (priv->new_term));

  // set the attributes of the new terminal
  priv->new_term.c_lflag &= ~ICANON;
  priv->new_term.c_lflag &= ~ECHO;
  priv->new_term.c_cc[VTIME] = 0; // latency
  priv->new_term.c_cc[VMIN] = 1; // nb char
  if (tcsetattr (STDIN_FILENO, TCSANOW, &(priv->new_term)) < 0)
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_PROGRAMMING,
        "%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_setup_terminal",
        "could not set term attributes");
    return FALSE;
  }
  return TRUE;
}

static void
ui_controller_term_restore_terminal (UiControllerTerm *controller_term, GError **error)
{
  UiControllerTermPrivate *priv = UI_CONTROLLER_TERM_GET_PRIVATE (controller_term);

  tcsetattr (STDIN_FILENO, TCSANOW, &(priv->old_term));
}

/**
 * ui_controller_term_configure_impl:
 * @controller:
 * @filename:
 * @error:
 * @returns: TRUE if the action was successful, FALSE otherwise
 *
 * This is the implementation of the class interface.
 **/
static gboolean
ui_controller_term_configure_impl (UiController *controller,
    const gchar *filename, GError **error)
{
  if (!UI_IS_CONTROLLER_TERM (controller))
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_PROGRAMMING,
        "%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_configure_impl", "invalid \'controller\' argument");
    return FALSE;
  }
  /* There is no default implementation, so return an error */
  GTimeVal timestamp;
  g_get_current_time (&timestamp);
  g_set_error (error,
      UI_CONTROLLER_TERM_ERROR,
      UI_CONTROLLER_TERM_ERROR_NOT_IMPLEMENTED,
      "%s<#>%s<#>%s",
      g_time_val_to_iso8601 (&timestamp),
      "ui_controller_term_configure_impl", "no implementation available");
  return FALSE;
}

/**
 * ui_controller_term_error_quark:
 * @returns: A GQuark
 *
 * Defines the error
 **/
GQuark
ui_controller_term_error_quark ()
{
  return g_quark_from_string ("UI_CONTROLLER_TERM_ERROR");
}

/**
 * ui_controller_term_add_stdin_source:
 * @controller_term:
 * @error:
 *
 * Create an input watch on stdin as a source of event for the main context.
 **/
static void
ui_controller_term_add_stdin_source (UiControllerTerm *controller_term, GError **error)
{
  if (!UI_IS_CONTROLLER_TERM (controller_term))
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_PROGRAMMING,
        "%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_add_stdin_source", "invalid \'controller\' argument");
    return;
  }

  GIOChannel *channel;
  GSource *source;

  UiControllerTermPrivate *priv = UI_CONTROLLER_TERM_GET_PRIVATE (controller_term);

  ui_controller_term_setup_terminal (controller_term, NULL);

  /* create an input watch on stdin */
  channel = g_io_channel_unix_new (STDIN_FILENO);

  /* the source will be dispatched when there is data available for reading. */
  source = g_io_create_watch (channel, G_IO_IN);
  g_source_set_callback (source, (GSourceFunc) ui_controller_term_handle_stdin,
      controller_term, NULL);
  /* TODO: What if main_context not set yet?
   * TODO: Can we check return value of following call? */
  priv->source_tag = g_source_attach (source, priv->main_context);
  g_source_unref (source);
}

/**
 * ui_controller_term_handle_stdin:
 * @channel:
 * @condition:
 * @data:
 * @returns: FALSE if the event source should be removed.
 *
 * Implements a GIOFunc
 * TODO: Need to be fixed... doesn't handle escape sequence correctly.
 **/
static gboolean
ui_controller_term_handle_stdin (GIOChannel *channel,
    GIOCondition condition, gpointer data)
{
  GError *err = NULL;
  g_return_val_if_fail (UI_IS_CONTROLLER_TERM (data), FALSE);

  UiControllerTerm *controller_term = UI_CONTROLLER_TERM (data);

  gchar character;
  gsize bytes_read;
  static gboolean escaped = FALSE;

  if (g_io_channel_read_chars (channel, &character, 1,
          &bytes_read, &err) == G_IO_STATUS_ERROR)
  {
    g_error ("An error occured: %s.\n", err->message);
    g_error_free (err);
    return FALSE;
  }
  if (character == 91) // [
  {
    g_message ("escaped = true.\n");
    escaped = TRUE;
  }
  else if (escaped) 
  {
    if (character >= 65 && character < 127)
    {
      if (character == 65) // Key Up
      {
        g_message ("key up\n");
        g_signal_emit (controller_term, 
            signals [UI_CONTROLLER_TERM_SIGNAL_KEY_UP], 0, NULL);
      } 
      else if (character == 66) // Key Down
      {
        g_message ("key down\n");
      }
      else if (character == 67)
      {
        g_message ("key right\n");
      }
      else if (character == 68)
      {
        g_message ("key left\n");
      }
      else
      {
        g_message ("unknown key %d.\n", character);
      }
    }
    else 
    {
      if (character == 51) // Key Del
      {
        g_message ("key delete\n");
      }
      else
      {
        g_message ("bizarre key: %d.\n", character);
      }
    }
    escaped = FALSE;
    g_message ("escape is now false\n");
  } // else if (escaped)
  else if (character >= 32 && character < 127) // Key ascii char
  {
    g_message ("ascii: %c\n", character);
  }
  else if (character == 127) // Key Backspace
  {
    g_message ("key backspace\n");
  }
  else if (character == 10) // Key Enter
  {
    g_message ("key enter\n");
  }
  else if (character == 8)
  {
    g_message ("key 8\n");
  }
  else if (character == 9)
  {
    g_message ("key tab\n");
  }
  else if (character == 27)
  {
    g_message ("key escape\n");
  }
  else 
  {
    g_print ("?? fin: %d.\n", character);
  }

  return TRUE;
}

/**
 * ui_controller_term_connect_with_main_loop_impl:
 * @controller:
 * @main_loop:
 * @main_context:
 * @error:
 * @returns: TRUE if the action was successful, FALSE otherwise
 **/
static gboolean
ui_controller_term_connect_with_main_loop_impl (UiController *controller,
    GMainLoop *main_loop, GMainContext *main_context, GError **error)
{
  GError *err = NULL;

  if (!UI_IS_CONTROLLER_TERM (controller))
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_PROGRAMMING,
        "%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_connect_with_main_loop_impl",
        "invalid \'controller\' argument");
    return FALSE;
  }

  UiControllerTerm *self = UI_CONTROLLER_TERM (controller);
  UiControllerTermPrivate *priv = UI_CONTROLLER_TERM_GET_PRIVATE (controller);

  priv->main_loop = main_loop;
  priv->main_context = main_context;

  ui_controller_term_add_stdin_source (self, &err);
  if (err != NULL)
  {
    g_propagate_error (error, err);
    return FALSE;
  }
  return TRUE;
}

/**
 * ui_controller_term_register_with_dbus_impl:
 * @controller:
 * @object_path:
 * @bus:
 * @error:
 * @returns: TRUE if the action was successful, FALSE otherwise
 **/
static gboolean
ui_controller_term_register_with_dbus_impl (UiController *controller,
    const gchar *object_path, DBusGConnection * bus, GError **error)
{
  if (!UI_IS_CONTROLLER_TERM (controller))
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_PROGRAMMING,
        "%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_register_with_dbus_impl", "invalid \'controller\' argument");
    return FALSE;
  }

  DBusGProxy *bus_proxy = dbus_g_proxy_new_for_name (bus,
      DBUS_SERVICE_DBUS, DBUS_PATH_DBUS,
      DBUS_INTERFACE_DBUS);
  if (bus_proxy == NULL)
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_DBUS,
        "%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_register_with_dbus_impl", "unable to retrieve D-Bus proxy");
    return FALSE;
  }
  guint result;
  GError *err = NULL;
  if (!dbus_g_proxy_call (bus_proxy, "RequestName", &err,
          G_TYPE_STRING, UI_CONTROLLER_TERM_SERVICE_NAME,
          G_TYPE_UINT, 0, G_TYPE_INVALID, G_TYPE_UINT, &result, G_TYPE_INVALID))
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_DBUS,
        "%s<#>%s<#>%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_register_with_dbus_impl",
        "unable to execute D-Bus RequestName RPC",
        UI_CONTROLLER_TERM_SERVICE_NAME, err->message);
    g_error_free (err);
    return FALSE;
  }
  if (result != 1)
  {
    GTimeVal timestamp;
    g_get_current_time (&timestamp);
    g_set_error (error,
        UI_CONTROLLER_TERM_ERROR,
        UI_CONTROLLER_TERM_ERROR_DBUS,
        "%s<#>%s<#>%s<#>%s<#>%s",
        g_time_val_to_iso8601 (&timestamp),
        "ui_controller_term_register_with_dbus_impl",
        "unable to get the primary well-known name",
        UI_CONTROLLER_TERM_SERVICE_NAME,
        err == NULL ? "no details available" : err->message);
    if (err != NULL)
      g_error_free (err);
    return FALSE;
  }

  /* TODO: Can we check success of the following  */
  dbus_g_connection_register_g_object (bus,
      UI_CONTROLLER_TERM_OBJECT_PATH, G_OBJECT (controller));

  return TRUE;
}
