#ifndef __UI_CONTROLLER_H__
#define __UI_CONTROLLER_H__

#include <glib-object.h>
#include <dbus/dbus-glib.h>

#include "ui-constants.h"
#include "ui-types.h"
#include "ui-enumtypes.h"

G_BEGIN_DECLS

#define UI_TYPE_CONTROLLER (ui_controller_get_type ())

#define UI_CONTROLLER(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST ((obj), \
  UI_TYPE_CONTROLLER, UiController))

#define UI_CONTROLLER_CONST(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST ((obj), \
  UI_TYPE_CONTROLLER, UiController const))

#define UI_CONTROLLER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST ((klass), \
  UI_TYPE_CONTROLLER, UiControllerClass))

#define UI_IS_CONTROLLER(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE ((obj), \
  UI_TYPE_CONTROLLER))

#define UI_IS_CONTROLLER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE ((klass), \
  UI_TYPE_CONTROLLER))

#define UI_CONTROLLER_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_CLASS ((obj), \
  UI_TYPE_CONTROLLER, UiControllerClass))

#define UI_CONTROLLER_ERROR ui_controller_error_quark ()

typedef enum
{
  UI_CONTROLLER_SIGNAL_CONFIGURED = 0,
  UI_CONTROLLER_SIGNAL_COUNT
} UiControllerSignal;

typedef enum
{
  UI_CONTROLLER_PROPERTY_UNKNOWN = 0,
  UI_CONTROLLER_PROPERTY_ID,
  UI_CONTROLLER_PROPERTY_COUNT
} UiControllerProperty;

typedef enum
{
  UI_CONTROLLER_ERROR_UNKNOWN = 0,
  UI_CONTROLLER_ERROR_PROGRAMMING,
  UI_CONTROLLER_ERROR_NOT_IMPLEMENTED,
  UI_CONTROLLER_ERROR_CONFIGURATION,
  UI_CONTROLLER_ERROR_COUNT
} UiControllerError;

struct _UiController
{
  GObject parent;
};

typedef struct _UiControllerClass UiControllerClass;

struct _UiControllerClass
{
  GObjectClass parent_class;

  gboolean (*connect_with_main_loop) (UiController *controller,
    GMainLoop *main_loop, GMainContext *main_context, GError **error);

  gboolean (*configure) (UiController *controller, const gchar *filename,
    GError **error);

  gboolean (*register_with_dbus) (UiController *controller, const gchar *object_path,
    DBusGConnection * bus, GError **error);

};

GType ui_controller_get_type (void) G_GNUC_CONST;

UiController *ui_controller_new (gchar *id);

void ui_controller_free (gpointer data);

gboolean ui_controller_configure (UiController *controller,
    const gchar *filename, GError **error);

gboolean ui_controller_connect_with_main_loop (UiController *controller,
    GMainLoop *main_loop, GMainContext *main_context, GError **error);

gboolean ui_controller_register_with_dbus (UiController *controller,
    const gchar *object_path, DBusGConnection * bus, GError **error);

const gchar *ui_controller_get_id (UiController *controller);

GQuark ui_controller_error_quark ();

G_END_DECLS
#endif /* __UI_CONTROLLER_H__ */
