#include <stdlib.h>

#include <glib.h>
#include <glib/gprintf.h>
#include <dbus/dbus-glib.h>

#include "ui.h"

int
main (int argc, char **argv)
{
  GError *error = NULL;
  
  g_type_init ();

  GMainContext *main_context = g_main_context_new ();
  GMainLoop *main_loop = g_main_loop_new (main_context, FALSE);

  DBusGConnection *bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
  if (error != NULL)
  {
    g_error ("Error connecting to Session D-Bus: %s", error->message);
    g_error_free (error);
    exit (-1);
  }

  UiController *controller = ui_controller_term_new ("Auction Sniper Controller");
  ui_controller_register_with_dbus (controller, UI_CONTROLLER_TERM_OBJECT_PATH, 
      bus, &error);
  if (error != NULL)
  {
    g_error ("Error registering controller with D-Bus: %s", error->message);
    g_error_free (error);
    exit (-1);
  }

  ui_controller_connect_with_main_loop (controller, main_loop, main_context, &error);
  if (error != NULL)
  {
    g_error ("Error registering controller with mainloop: %s", error->message);
    g_error_free (error);
    exit (-1);
  }

  g_main_loop_run (main_loop);

  ui_controller_free (controller);
  return 0;
}
